
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <time.h>

#include <SDL/SDL.h>
#include <SDL/SDL_video.h>
#include <SDL/SDL_keysym.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "defines.h"
#include "main.h"
#include "txt.h"
#include "lsload.h"
#include "screen_capture.h"


extern int dirty_textures;


const SDL_VideoInfo *sdl_video_info = NULL;

GLfloat light_ambient[] = { 10, 10, 10, 1.0 };

struct particle_system *particle_systems = NULL;
struct object_object obj_kp;
struct object_mesh *meshes_first = NULL, *meshes_last = NULL;
int lsload_errors = ON, quit = NO;

float camx, camy, camz, desx, desy, desz;
float mirnx, mirny, mirnz;
float mirror[4*3] = { -1.0, 0.5, -1.5,
		      -1.0, 1.5, -1.5,
		       1.0, 1.5, -1.5,
		       1.0, 0.5, -1.5 };

int key_status = 0;
int mirror_txt_id, particle_txt_id, dirt_txt_id;

/* a, b, c, area */
float mirror_emitter_1[4*3+1], mirror_emitter_2[4*3+1];

/* particle system */
int particle_systems_id = 0;

/* particle sorting list */
struct particle **particle_list = NULL;
int particle_list_size = 0;


int main(int argc, char *argv[]) {

  struct particle_system *p;
  int i;


  for (i = 0; i < 3*3; i++)
    mirror_emitter_1[i] = mirror[i];
  for (i = 0; i < 3*3; i++)
    mirror_emitter_2[i] = mirror[(i+6)%12];

  srand(time(NULL));

  particles_add(PARTICLE_SYSTEM_POINT_BLOOD, -0.5, 0.97, -0.02, -1, 0, 0, NULL, &i);
  particles_add(PARTICLE_SYSTEM_POINT_BLOOD, 0.3, 0.97, -0.02, 1, 0, 0, NULL, &i);
  particles_add(PARTICLE_SYSTEM_POINT_FLAME, 0, 1.4, 0, 0.5, 1, 0.2, NULL, &i);
  particles_add(PARTICLE_SYSTEM_POINT_FLAME, -0.06, 1.4, -0.01, -1, 1, -0.2, NULL, &i);
  particles_add(PARTICLE_SYSTEM_POINT_FLAME, -0.08, 1.4, 0.04, -0.3, 0.7, 1, NULL, &i);
  particles_add(PARTICLE_SYSTEM_FACE_MIST, 0, 0, 0, 0, 0, 1, mirror_emitter_1, &i);
  particles_add(PARTICLE_SYSTEM_FACE_MIST, 0, 0, 0, 0, 0, 1, mirror_emitter_2, &i);

  particles_add(PARTICLE_SYSTEM_POINT_FLAME, 0, 1.4, 3, 0, 0, 1, NULL, &i);
  particles_get(i, &p);
 
  txt_init("data/txt/");
  opengl_init();
  objects_init();
  particles_init();

  txt_create("mirror", &mirror_txt_id, 256, 256);
  txt_load("particle-a.s3b", &particle_txt_id);
  txt_load("dirt.s3b", &dirt_txt_id);

  /* camera */
  camx = 0.2;
  camy = 2;
  camz = 4;
  desx = 0;
  desy = 1.3;
  desz = 0;

  /* mirror */
  mirnx = 0;
  mirny = 0;
  mirnz = 1;

  SDL_ShowCursor(SDL_DISABLE);
  SDL_SetEventFilter(sdl_filter_events);

  while (quit == NO) {
    keyboard_io();

    if (key_status & KEY_STATUS_LEFT)
      camx -= 0.02;
    if (key_status & KEY_STATUS_RIGHT)
      camx += 0.02;
    if (key_status & KEY_STATUS_DOWN)
      camz += 0.02;
    if (key_status & KEY_STATUS_UP)
      camz -= 0.02;

    p->z -= 0.01;
    if (p->z <= -1.5) {
      p->x = rand()/(float)RAND_MAX * 2 - 1;
      p->y = rand()/(float)RAND_MAX * 1 + 1;
      p->z = 2;
    }

    opengl_render_scene();
  }

  SDL_ShowCursor(SDL_ENABLE);
  SDL_Quit();

  return 0;
}


int opengl_init(void) {

  if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_TIMER) < 0) {
    fprintf(stderr, "INIT_OPENGL: Video initialization failed: %s\n", SDL_GetError());
    return FAILED;
  }

  sdl_video_info = SDL_GetVideoInfo();

  SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 8);
  SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 8);
  SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 8);
  SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, 16);
  SDL_GL_SetAttribute(SDL_GL_STENCIL_SIZE, 8);
  SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);

  if (SDL_SetVideoMode(SCREEN_X, SCREEN_Y, sdl_video_info->vfmt->BitsPerPixel, SDL_OPENGL | SDL_FULLSCREEN) == 0) {
    fprintf(stderr, "INIT_OPENGL: Video mode initialization failed: %s\n", SDL_GetError());
    return FAILED;
  }

  SDL_EnableUNICODE(1);

  return SUCCEEDED;
}


int objects_init(void) {

  struct object_mesh *m;


  meshes_first = NULL;
  meshes_last = NULL;

  if (object_load_landscape("data/obj/church.lan") == FAILED)
    return FAILED;

  obj_kp.meshes = meshes_first;
  obj_kp.display_list = 0;
  obj_kp.next = NULL;
  obj_kp.x = 0;
  obj_kp.y = 0;
  obj_kp.z = 0;

  m = obj_kp.meshes;
  while (m != NULL) {
    mesh_complete(m);
    m = m->next;
  }

  meshes_first = NULL;
  meshes_last = NULL;

  return SUCCEEDED;
}


int opengl_render_scene(void) {

  struct txt_object *txt;
  GLdouble cp[4] = { 0, 0, 1, 0 };
  GLfloat mat[4*4];
  float x, y, z, k;


  x = (mirror[3*0 + 0] + mirror[3*1 + 0] + mirror[3*2 + 0] + mirror[3*3 + 0])/4;
  y = (mirror[3*0 + 1] + mirror[3*1 + 1] + mirror[3*2 + 1] + mirror[3*3 + 1])/4;
  z = (mirror[3*0 + 2] + mirror[3*1 + 2] + mirror[3*2 + 2] + mirror[3*3 + 2])/4;

  particles_update();

  glViewport(0, 0, SCREEN_X, SCREEN_Y);
  glClearColor(0, 0, 0, 0);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
  glShadeModel(GL_SMOOTH);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
  glDepthFunc(GL_LESS);
  glFrontFace(GL_CW);
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, light_ambient);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(42, (double)SCREEN_X / SCREEN_Y, 0.1, 10);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluLookAt(camx, camy, camz, desx, desy, desz, 0, 1, 0);

  glEnable(GL_TEXTURE_2D);
  glEnable(GL_CULL_FACE);
  glEnable(GL_LIGHTING);
  glEnable(GL_DEPTH_TEST);

  glPushMatrix();

  /* reflection transformation matrix */
  k = mirror[3*0 + 0]*mirnx + mirror[3*0 + 1]*mirny + mirror[3*0 + 2]*mirnz;

  mat[0] = 1 - 2*mirnx*mirnx;
  mat[1] = -2*mirnx*mirny;
  mat[2] = -2*mirnx*mirnz;
  mat[3] = 0;

  mat[4] = -2*mirnx*mirny;
  mat[5] = 1 - 2*mirnx*mirnx;
  mat[6] = -2*mirny*mirnz;
  mat[7] = 0;

  mat[8] = -2*mirnx*mirnz;
  mat[9] = -2*mirny*mirnz;
  mat[10] = 1 - 2*mirnz*mirnz;
  mat[11] = 0;

  mat[12] = 2*k*mirnx;
  mat[13] = 2*k*mirny;
  mat[14] = 2*k*mirnz;
  mat[15] = 1;

  glMultMatrixf(mat);

  /* clipping plane for clipping the part of the scene away that's beyond the mirror plane */
  cp[3] = -mirror[3*0 + 2];
  glClipPlane(GL_CLIP_PLANE0, cp);
  glEnable(GL_CLIP_PLANE0);

  /* DRAW THE SCENE - PASS 1 (MIRROR) */
  glTranslatef(obj_kp.x, obj_kp.y, obj_kp.z);
  render_object(&obj_kp);

  particles_draw();

  glDisable(GL_CLIP_PLANE0);

  glPopMatrix();
  glFrontFace(GL_CCW);

  /* second pass */
  glClear(GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

  glStencilOp(GL_REPLACE, GL_REPLACE, GL_REPLACE);
  glStencilFunc(GL_ALWAYS, 1, 1);
  glEnable(GL_STENCIL_TEST);
  glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);

  /* draw the mirror */
  glDisable(GL_TEXTURE_2D);
  glBegin(GL_TRIANGLES);

  glVertex3f(mirror[3*0 + 0], mirror[3*0 + 1], mirror[3*0 + 2]);
  glVertex3f(mirror[3*2 + 0], mirror[3*2 + 1], mirror[3*2 + 2]);
  glVertex3f(mirror[3*1 + 0], mirror[3*1 + 1], mirror[3*1 + 2]);

  glVertex3f(mirror[3*0 + 0], mirror[3*0 + 1], mirror[3*0 + 2]);
  glVertex3f(mirror[3*3 + 0], mirror[3*3 + 1], mirror[3*3 + 2]);
  glVertex3f(mirror[3*2 + 0], mirror[3*2 + 1], mirror[3*2 + 2]);

  glEnd();

  /* remove data outside the mirror area */
  glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
  glStencilFunc(GL_NOTEQUAL, 1, 1);
  glDisable(GL_DEPTH_TEST);
  glDisable(GL_LIGHTING);
  glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);

  /* draw a full screen quad */
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glColor4f(0, 0, 0, 0);

  glBegin(GL_QUADS);
  glVertex3i(-1, -1, -1);
  glVertex3i(1, -1, -1);
  glVertex3i(1, 1, -1);
  glVertex3i(-1, 1, -1);
  glEnd();

  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  glPopMatrix();

  glDisable(GL_STENCIL_TEST);

  /* draw the mirror's second layer (dirty mirror) */
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_BLEND);
  glEnable(GL_TEXTURE_2D);
  glColor4f(0.5, 0.5, 0.5, 0.5);
  txt_get(dirt_txt_id, &txt);
  glBindTexture(GL_TEXTURE_2D, txt->ogl_name);

  glBegin(GL_QUADS);

  glTexCoord2f(0, 1);
  glVertex3f(mirror[3*2 + 0], mirror[3*2 + 1], mirror[3*2 + 2]);
  glTexCoord2f(1, 1);
  glVertex3f(mirror[3*1 + 0], mirror[3*1 + 1], mirror[3*1 + 2]);
  glTexCoord2f(1, 0);
  glVertex3f(mirror[3*0 + 0], mirror[3*0 + 1], mirror[3*0 + 2]);
  glTexCoord2f(0, 0);
  glVertex3f(mirror[3*3 + 0], mirror[3*3 + 1], mirror[3*3 + 2]);

  glEnd();

  glDisable(GL_BLEND);
  glEnable(GL_LIGHTING);
  glEnable(GL_DEPTH_TEST);

  /* DRAW THE SCENE - PASS 2 */
  glPushMatrix();
  glTranslatef(obj_kp.x, obj_kp.y, obj_kp.z);
  render_object(&obj_kp);
  glPopMatrix();

  particles_draw();

  glDisable(GL_TEXTURE_2D);
  glDisable(GL_LIGHTING);
  glDisable(GL_CULL_FACE);
  glDisable(GL_DEPTH_TEST);

  SDL_GL_SwapBuffers();

  return SUCCEEDED;
}


int mesh_complete(struct object_mesh *mesh) {

  double dx1, dx2, dy1, dy2, dz1, dz2, r1, r2, r3, x1, x2, x3, y1, y2, y3, z1, z2, z3, *t;
  float *f;
  int i, *tl;


  if (mesh == NULL)
    return FAILED;

  /* generate missing normals */
  if (mesh->normals == 0) {

    mesh->normals = mesh->triangles;
    mesh->normal_list = malloc(sizeof(float) * mesh->triangles * 3);
    if (mesh->normal_list == NULL) {
      fprintf(stderr, "MESH_COMPLETE: Out of memory.\n");
      return FAILED;
    }

    tl = mesh->triangle_normal_list;
    for (i = 0; i < mesh->triangles; i++) {
      *(tl++) = i;
      *(tl++) = i;
      *(tl++) = i;
    }

    /* compute triangle plane normals */
    tl = mesh->triangle_list;
    f = mesh->normal_list;
    t = mesh->vertice_list;
    for (i = 0; i < mesh->triangles; i++) {
      x1 = t[(*tl) * 3];
      y1 = t[(*tl) * 3 + 1];
      z1 = t[(*tl) * 3 + 2];
      x2 = t[(*(tl + 1)) * 3];
      y2 = t[(*(tl + 1)) * 3 + 1];
      z2 = t[(*(tl + 1)) * 3 + 2];
      x3 = t[(*(tl + 2)) * 3];
      y3 = t[(*(tl + 2)) * 3 + 1];
      z3 = t[(*(tl + 2)) * 3 + 2];
      tl += 3;

      dx1 = x1 - x2;
      dy1 = y1 - y2;
      dz1 = z1 - z2;
      dx2 = x1 - x3;
      dy2 = y1 - y3;
      dz2 = z1 - z3;
      r1 = dy1*dz2 - dy2*dz1;
      r2 = dx2*dz1 - dx1*dz2;
      r3 = dx1*dy2 - dx2*dy1;

      dx1 = r1*r1 + r2*r2 + r3*r3;
      dx2 = sqrt(dx1);
      r1 /= dx2;
      r2 /= dx2;
      r3 /= dx2;

      *(f++) = r1;
      *(f++) = r2;
      *(f++) = r3;
    }
  }

  return SUCCEEDED;
}


int render_object(struct object_object *o) {

  struct object_mesh *s;
  struct txt_object *tx;
  double *d;
  float *l, *q;
  int i, *a, id, *k, *t, h;


  /* do we have a display list of the object? */
  if (o->display_list != 0) {
    glCallList(o->display_list);
    return SUCCEEDED;
  }

  s = o->meshes;

  id = -1;
  while (s != NULL) {
    i = s->triangles;
    a = s->triangle_list;
    d = s->vertice_list;
    q = s->normal_list;
    l = s->uv_list;
    k = s->triangle_normal_list;
    t = s->textures;

    /* txt not loaded yet? */
    if (id != *t) {
      if (txt_get(*t, &tx) == SUCCEEDED) {
	glBindTexture(GL_TEXTURE_2D, tx->ogl_name);
	/* do we have translucent triangles? */
	if (tx->status & TXT_STATUS_ALPHA) {
	  glEnable(GL_BLEND);
	  glDisable(GL_CULL_FACE);
	  glColor4f(1, 1, 1, tx->alpha);
	}
	else {
	  glDisable(GL_BLEND);
	  glEnable(GL_CULL_FACE);
	}
	id = *t;
      }
    }

    glBegin(GL_TRIANGLES);
    while (i > 0) {
      /* txt not loaded yet? */
      if (id != *t) {
	if (txt_get(*t, &tx) == SUCCEEDED) {
	  glEnd();
	  glBindTexture(GL_TEXTURE_2D, tx->ogl_name);
	  /* do we have translucent triangles? */
	  if (tx->status & TXT_STATUS_ALPHA) {
	    glEnable(GL_BLEND);
	    glColor4f(1, 1, 1, tx->alpha);
	  }
	  else
	    glDisable(GL_BLEND);
	  glBegin(GL_TRIANGLES);
	  id = *t;
	}
      }
      t++;
      glTexCoord2f(*l, *(l+1));
      h = *(k++) * 3;
      glNormal3f(q[h + 0], q[h + 1], q[h + 2]);
      h = *(a++) * 3;
      glVertex3f(d[h + 0], d[h + 1], d[h + 2]);
      l += 2;
      glTexCoord2f(*l, *(l+1));
      h = *(k++) * 3;
      glNormal3f(q[h + 0], q[h + 1], q[h + 2]);
      h = *(a++) * 3;
      glVertex3f(d[h + 0], d[h + 1], d[h + 2]);
      l += 2;
      glTexCoord2f(*l, *(l+1));
      h = *(k++) * 3;
      glNormal3f(q[h + 0], q[h + 1], q[h + 2]);
      h = *(a++) * 3;
      glVertex3f(d[h + 0], d[h + 1], d[h + 2]);
      l += 2;
      i--;
    }
    glEnd();
    s = s->next;
  }

  return SUCCEEDED;
}


int keyboard_io(void) {

  SDL_Event event;
  SDLMod mod;
  int i;


  while (SDL_PollEvent(&event)) {
    switch (event.type) {

    /* mouse */
    case SDL_MOUSEBUTTONDOWN:
      /* pick up */
      if (event.button.button == 3)
	;
      break;

    /* key up */
    case SDL_KEYUP:
      i = event.key.keysym.sym;
      switch (event.key.keysym.sym) {
      case SDLK_i:
	SDL_WM_IconifyWindow();
	break;
      case SDLK_LEFT:
      case SDLK_a:
	key_status &= ~KEY_STATUS_LEFT;
	break;
      case SDLK_RIGHT:
      case SDLK_d:
	key_status &= ~KEY_STATUS_RIGHT;
	break;
      case SDLK_UP:
      case SDLK_w:
	key_status &= ~KEY_STATUS_UP;
	break;
      case SDLK_DOWN:
      case SDLK_s:
	key_status &= ~KEY_STATUS_DOWN;
	break;
      case SDLK_ESCAPE:
	quit = YES;
	break;
      default:
	break;
      }
      break;

    /* key down */
    case SDL_KEYDOWN:
      i = event.key.keysym.sym;
      switch (event.key.keysym.sym) {
      case SDLK_i:
	SDL_WM_IconifyWindow();
	break;
      case SDLK_LEFT:
      case SDLK_a:
	key_status |= KEY_STATUS_LEFT;
	break;
      case SDLK_RIGHT:
      case SDLK_d:
	key_status |= KEY_STATUS_RIGHT;
	break;
      case SDLK_UP:
      case SDLK_w:
	key_status |= KEY_STATUS_UP;
	break;
      case SDLK_DOWN:
      case SDLK_s:
	key_status |= KEY_STATUS_DOWN;
	break;
      case SDLK_ESCAPE:
	quit = YES;
	break;
      case SDLK_F12:
	screen_capture();
	break;
      default:
	break;
      }
      break;
    }
  }

  mod = SDL_GetModState();

  return SUCCEEDED;
}


int sdl_filter_events(const SDL_Event *event) {

  int player_x = 0, player_y = 0;


  if (event->type == SDL_QUIT)
    return 1;

  if (event->type == SDL_MOUSEMOTION) {
    player_x += event->motion.xrel / (double)SCREEN_X * 360;
    player_y += event->motion.yrel / (double)SCREEN_Y * 360;
    if (player_x < 0)
      player_x += 360;
    if (player_x >= 360)
      player_x -= 360;
    if (player_y < -70)
      player_y = -70;
    if (player_y > 70)
      player_y = 70;
  }

  return 1;
}


int particles_init(void) {

  return SUCCEEDED;
}


int _particles_sort(const void *a, const void *b) {

  struct particle *c = *(struct particle **)a, *d = *(struct particle **)b;

  if (c->d > d->d)
    return -1;
  return 1;
}


void _particles_draw(struct particle **pl, float *bbv, int n) {

  GLfloat c[4*3], texcoords[4*2];
  float a;
  int i;


  glEnableClientState(GL_TEXTURE_COORD_ARRAY);
  glEnableClientState(GL_VERTEX_ARRAY);

  glTexCoordPointer(2, GL_FLOAT, 0, texcoords);
  glVertexPointer(3, GL_FLOAT, 0, c);

  texcoords[0*2 + 0] = 1;
  texcoords[0*2 + 1] = 1;
  texcoords[1*2 + 0] = 0;
  texcoords[1*2 + 1] = 1;
  texcoords[2*2 + 0] = 0;
  texcoords[2*2 + 1] = 0;
  texcoords[3*2 + 0] = 1;
  texcoords[3*2 + 1] = 0;

  for (i = 0; i < n; i++) {
    if ((pl[i])->age > 0) {

      a = 0.02;

      c[2*3 + 0] = (pl[i])->x + bbv[0]*(pl[i])->sc;
      c[2*3 + 1] = (pl[i])->y + bbv[1]*(pl[i])->sc;
      c[2*3 + 2] = (pl[i])->z + bbv[2]*(pl[i])->sc;
      c[3*3 + 0] = (pl[i])->x + bbv[3]*(pl[i])->sc;
      c[3*3 + 1] = (pl[i])->y + bbv[4]*(pl[i])->sc;
      c[3*3 + 2] = (pl[i])->z + bbv[5]*(pl[i])->sc;
      c[0*3 + 0] = (pl[i])->x + bbv[6]*(pl[i])->sc;
      c[0*3 + 1] = (pl[i])->y + bbv[7]*(pl[i])->sc;
      c[0*3 + 2] = (pl[i])->z + bbv[8]*(pl[i])->sc;
      c[1*3 + 0] = (pl[i])->x + bbv[9]*(pl[i])->sc;
      c[1*3 + 1] = (pl[i])->y + bbv[10]*(pl[i])->sc;
      c[1*3 + 2] = (pl[i])->z + bbv[11]*(pl[i])->sc;

      a = (50.0-(pl[i])->age)/100;
      if (a < 0) {
	(pl[i])->age = 0;
	a = 0;
      }

      glColor4f((pl[i])->cr, (pl[i])->cg, (pl[i])->cb, a);

      glDrawArrays(GL_QUADS, 0, 4);
    }
  }

  glDisableClientState(GL_VERTEX_ARRAY);
  glDisableClientState(GL_TEXTURE_COORD_ARRAY);
}


void _particles_compute_face_area(struct particle_system *pa) {

  float *f, d, e, dx, dy, dz, ex, ey, ez, a;


  f = pa->tri;

  /* compute the area of the triangle */
  dx = f[0*3 + 0] - f[1*3 + 0];
  dy = f[0*3 + 1] - f[1*3 + 1];
  dz = f[0*3 + 2] - f[1*3 + 2];
  d = sqrt(dx*dx + dy*dy + dz*dz);
  ex = f[0*3 + 0] - f[2*3 + 0];
  ey = f[0*3 + 1] - f[2*3 + 1];
  ez = f[0*3 + 2] - f[2*3 + 2];
  e = sqrt(ex*ex + ey*ey + ez*ez);

  a = acos((dx*ex + dy*ey + dz*ez)/(d*e));
  f[4*3] = sin(a)*0.5*d*e;
}


int _point_inside_triangle_f(float *p, float *t) {

  float u[3], v[3], w[3], uu, uv, vv, wu, wv, s, n, d;


  /* u ~ 0->1 */
  u[0] = t[3*1 + 0] - t[3*0 + 0];
  u[1] = t[3*1 + 1] - t[3*0 + 1];
  u[2] = t[3*1 + 2] - t[3*0 + 2];

  /* v ~ 0->2 */
  v[0] = t[3*2 + 0] - t[3*0 + 0];
  v[1] = t[3*2 + 1] - t[3*0 + 1];
  v[2] = t[3*2 + 2] - t[3*0 + 2];

  /* w ~ 0->p */
  w[0] = p[0] - t[3*0 + 0];
  w[1] = p[1] - t[3*0 + 1];
  w[2] = p[2] - t[3*0 + 2];

  uu = u[0]*u[0] + u[1]*u[1] + u[2]*u[2];
  uv = u[0]*v[0] + u[1]*v[1] + u[2]*v[2];
  vv = v[0]*v[0] + v[1]*v[1] + v[2]*v[2];
  wu = w[0]*u[0] + w[1]*u[1] + w[2]*u[2];
  wv = w[0]*v[0] + w[1]*v[1] + w[2]*v[2];

  d = uv*uv - uu*vv;

  s = (uv*wv - vv*wu)/d;
  if (s < 0.0 || s > 1.0)
    return NO;

  n = (uv*wu - uu*wv)/d;
  if (n < 0.0 || (s+n) > 1.0)
    return NO;

  return YES;
}


int _particles_random_point_on_face(struct particle_system *pa) {

  float *f, p[3];
  float a, d, dx, dy, dz, ax, ay, az;


  f = pa->tri;

  while (1) {
    d = rand()/(double)RAND_MAX;
    a = rand()/(double)RAND_MAX;

    /* scale the edge a->c by a random value [0,1] */
    dx = f[2*3 + 0] - f[1*3 + 0];
    dy = f[2*3 + 1] - f[1*3 + 1];
    dz = f[2*3 + 2] - f[1*3 + 2];
    dx *= d;
    dy *= d;
    dz *= d;

    /* scale the edge a->b by a random value [0,1] */
    ax = f[0*3 + 0] - f[1*3 + 0];
    ay = f[0*3 + 1] - f[1*3 + 1];
    az = f[0*3 + 2] - f[1*3 + 2];
    ax *= a;
    ay *= a;
    az *= a;

    p[0] = f[1*3 + 0]+ dx + ax;
    p[1] = f[1*3 + 1]+ dy + ay;
    p[2] = f[1*3 + 2]+ dz + az;

    if (_point_inside_triangle_f(p, f) == YES)
      break;
  }

  f[3*3 + 0] = p[0];
  f[3*3 + 1] = p[1];
  f[3*3 + 2] = p[2];

  return SUCCEEDED;
}


int particles_draw(void) {

  struct txt_object *txt;
  GLint bl, li, te, dm, ff;
  GLfloat m[4*4];
  float up[3], right[3], bbv[4*3];


  txt_get(particle_txt_id, &txt);
  glBindTexture(GL_TEXTURE_2D, txt->ogl_name);

  glGetIntegerv(GL_BLEND, &bl);
  glGetIntegerv(GL_FRONT_FACE, &ff);
  glGetIntegerv(GL_LIGHTING, &li);
  glGetIntegerv(GL_TEXTURE_2D, &te);
  glGetIntegerv(GL_DEPTH_WRITEMASK, &dm);

  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable(GL_BLEND);
  glEnable(GL_TEXTURE_2D);
  glDisable(GL_LIGHTING);
  glDepthMask(GL_FALSE);
  glFrontFace(GL_CCW);

  /* precalc the billboard vectors and values */
  glGetFloatv(GL_MODELVIEW_MATRIX, m);

  right[0] = m[0];
  right[1] = m[4];
  right[2] = m[8];
  up[0] = m[1];
  up[1] = m[5];
  up[2] = m[9];

  bbv[0] = -right[0] - up[0];
  bbv[1] = -right[1] - up[1];
  bbv[2] = -right[2] - up[2];
  bbv[3] = right[0] - up[0];
  bbv[4] = right[1] - up[1];
  bbv[5] = right[2] - up[2];
  bbv[6] = right[0] + up[0];
  bbv[7] = right[1] + up[1];
  bbv[8] = right[2] + up[2];
  bbv[9] = up[0] - right[0];
  bbv[10] = up[1] - right[1];
  bbv[11] = up[2] - right[2];

  _particles_draw(particle_list, bbv, particle_list_size);

  /* return old opengl status */
  glDepthMask(dm);
  glFrontFace(ff);
  if (bl == GL_FALSE)
    glDisable(GL_BLEND);
  if (li == GL_TRUE)
    glEnable(GL_LIGHTING);
  if (te == GL_FALSE)
    glDisable(GL_TEXTURE_2D);

  return SUCCEEDED;
}


int particles_remove(int id) {

  struct particle_system *p;
  int a, b;


  p = particle_systems;
  while (p != NULL) {
    if (p->id == id) {
      /* free the particle system and reinit the particle list */
      particle_list_size -= p->amount;
      free(particle_list);

      particle_list = malloc(sizeof(struct particle *) * particle_list_size);
      if (particle_list == NULL) {
	particle_list_size = 0;
	fprintf(stderr, "PARTICLES_REMOVE: Out of memory error.\n");
	return FAILED;
      }

      /* remove the system from the system list */
      if (p->prev != NULL)
	p->prev->next = p->next;
      if (p->next != NULL)
	p->next->prev = p->prev;
      if (particle_systems == p)
	particle_systems = p->next;

      /* reinit the particle list */
      p = particle_systems;
      b = 0;
      while (p != NULL) {
	for (a = 0; a < p->amount; a++, b++)
	  particle_list[b] = &(p->particles[a]);
	p = p->next;
      }

      return SUCCEEDED;
    }
    p = p->next;
  }

  return FAILED;
}


int particles_add(int type, float x, float y, float z, float dx, float dy, float dz, float *tri, int *id) {

  struct particle_system *p;
  float d;
  int a, b;


  p = malloc(sizeof(struct particle_system));
  if (p == NULL) {
    fprintf(stderr, "PARTICLES_ADD: Out of memory error.\n");
    return FAILED;
  }

  /* normalize the direction vector */
  d = sqrt(dx*dx + dy*dy + dz*dz);
  dx /= d;
  dy /= d;
  dz /= d;

  p->type = type;
  p->x = x;
  p->y = y;
  p->z = z;
  p->dx = dx;
  p->dy = dy;
  p->dz = dz;
  p->tri = tri;
  p->prev = NULL;
  p->id = particle_systems_id++;
  *id = p->id;

  if (tri != NULL)
    _particles_compute_face_area(p);

  /* amount of max particles */
  if (type == PARTICLE_SYSTEM_POINT_FLAME)
    p->amount = PARTICLES_MAX;
  else if (type == PARTICLE_SYSTEM_POINT_BLOOD)
    p->amount = PARTICLES_MAX;
  else if (type == PARTICLE_SYSTEM_FACE_MIST) {
    /* amount = c*area */
    p->amount = 1024*1*p->tri[4*3];
  }
  else {
    fprintf(stderr, "PARTICLES_ADD: Unknown particle system type \"%d\".\n", type);
    return FAILED;
  }

  p->particles = malloc(sizeof(struct particle) * p->amount);
  if (p->particles == NULL) {
    fprintf(stderr, "PARTICLES_ADD: Out of memory error.\n");
    return FAILED;
  }

  /* clear the particles */
  for (a = 0; a < p->amount; a++)
    p->particles[a].age = 0;

  /* add the new system to the list of systems */
  p->next = particle_systems;
  if (particle_systems != NULL)
    particle_systems->prev = p;
  particle_systems = p;

  /* add the new particles to the particle sort list */
  if (particle_list == NULL)
    particle_list = malloc(sizeof(struct particle *) * p->amount);
  else
    particle_list = realloc(particle_list, sizeof(struct particle *) * (particle_list_size + p->amount));

  if (particle_list == NULL) {
    fprintf(stderr, "PARTICLES_ADD: Out of memory error.\n");
    return FAILED;
  }

  for (a = 0, b = particle_list_size; a < p->amount; a++, b++)
    particle_list[b] = &(p->particles[a]);

  particle_list_size += p->amount;

  return SUCCEEDED;
}


int particles_update(void) {

  struct particle_system *pa;
  struct particle *p;
  double dx, dy, dz, r;
  int i, d, z;


  pa = particle_systems;
  while (pa != NULL) {
    p = pa->particles;

    /* count the dead */
    for (d = 0, i = 0; i < pa->amount; i++) {
      if (p[i].age == 0)
	d++;
    }

    /* POINT BLOOD */
    if (pa->type == PARTICLE_SYSTEM_POINT_BLOOD) {

      /* wake up a few */
      if (d > 0) {
	for (z = 0, i = 0; i < pa->amount; i++) {
	  if (p[i].age == 0) {

	    /* compute a random direction */
	    dx = rand()/(double)RAND_MAX-0.5;
	    dy = rand()/(double)RAND_MAX-0.5;
	    dz = rand()/(double)RAND_MAX-0.5;
	    r = sqrt(dx*dx + dy*dy + dz*dz);
	    dx /= r;
	    dy /= r;
	    dz /= r;
	    r = dx*pa->dx + dy*pa->dy + dz*pa->dz;
	    dx = (dx - r*pa->dx)*0.002 + pa->dx*0.01;
	    dy = (dy - r*pa->dy)*0.002 + pa->dy*0.01;
	    dz = (dz - r*pa->dz)*0.002 + pa->dz*0.01;

	    p[i].age = 1;
	    p[i].x = pa->x;
	    p[i].y = pa->y;
	    p[i].z = pa->z;
	    p[i].vx = dx;
	    p[i].vy = dy;
	    p[i].vz = dz;
	    p[i].ax = 0;
	    p[i].ay = -0.002;
	    p[i].az = 0;
	    p[i].cr = 0.5;
	    p[i].cg = 0;
	    p[i].cb = 0;
	    p[i].sc = 0.02;
	    z++;
	    d--;
	    if (d == 0)
	      break;
	    if (z == 10)
	      break;
	  }
	}
      }
    }
    /* POINT FLAME */
    else if (pa->type == PARTICLE_SYSTEM_POINT_FLAME) {

      /* wake up a few */
      if (d > 0) {
	for (z = 0, i = 0; i < pa->amount; i++) {
	  if (p[i].age == 0) {

	    /* compute a random direction */
	    dx = rand()/(double)RAND_MAX-0.5;
	    dy = rand()/(double)RAND_MAX-0.5;
	    dz = rand()/(double)RAND_MAX-0.5;
	    r = sqrt(dx*dx + dy*dy + dz*dz);
	    dx /= r;
	    dy /= r;
	    dz /= r;
	    r = dx*pa->dx + dy*pa->dy + dz*pa->dz;
	    dx = (dx - r*pa->dx)*0.0006 + pa->dx*0.001;
	    dy = (dy - r*pa->dy)*0.0006 + pa->dy*0.001;
	    dz = (dz - r*pa->dz)*0.0006 + pa->dz*0.001;

	    p[i].age = 1;
	    p[i].x = pa->x;
	    p[i].y = pa->y;
	    p[i].z = pa->z;
	    p[i].vx = dx;
	    p[i].vy = dy;
	    p[i].vz = dz;
	    p[i].ax = pa->dx * 0.0002;
	    p[i].ay = pa->dy * 0.0002;
	    p[i].az = pa->dz * 0.0002;
	    p[i].cr = 0.4;
	    p[i].cg = 0;
	    p[i].cb = 0.5;
	    if ((rand() % 10) == 0) {
	      p[i].cr = 1;
	      p[i].cg = 1;
	      p[i].cb = 1;
	    }
	    p[i].sc = 0.01;
	    z++;
	    d--;
	    if (d == 0)
	      break;
	    if (z == 10)
	      break;
	  }
	}
      }
    }
    /* FACE MIST */
    else if (pa->type == PARTICLE_SYSTEM_FACE_MIST) {

      /* wake up a few */
      if (d > 0) {
	for (z = 0, i = 0; i < pa->amount; i++) {
	  if (p[i].age == 0) {

	    _particles_random_point_on_face(pa);

	    p[i].age = 1;
	    p[i].x = pa->tri[3*3 + 0];
	    p[i].y = pa->tri[3*3 + 1];
	    p[i].z = pa->tri[3*3 + 2];
	    p[i].vx = 0;
	    p[i].vy = 0;
	    p[i].vz = 0;
	    p[i].ax = pa->dx * 0.0001;
	    p[i].ay = pa->dy * 0.0001;
	    p[i].az = pa->dz * 0.0001;
	    p[i].cr = 1;
	    p[i].cg = 1;
	    p[i].cb = 0;
	    p[i].sc = 0.005;
	    z++;
	    d--;
	    if (d == 0)
	      break;
	    if (z >= (int)(20*1*pa->tri[4*3]))
	      break;
	  }
	}
      }
    }

    /* advance the living particles */
    for (i = 0; i < pa->amount; i++) {
      if (p[i].age > 0) {
	p[i].age += 1;
	p[i].x += p[i].vx;
	p[i].y += p[i].vy;
	p[i].z += p[i].vz;
	p[i].vx += p[i].ax;
	p[i].vy += p[i].ay;
	p[i].vz += p[i].az;
	/* compute particle distance to the viewer */
	dx = camx - p[i].x;
	dy = camy - p[i].y;
	dz = camz - p[i].z;
	p[i].d = dx*dx + dy*dy + dz*dz;
      }
    }

    pa = pa->next;
  }

  /* sort the particles (distance to the camera) */
  qsort(particle_list, particle_list_size, sizeof(struct particle *), _particles_sort);

  return SUCCEEDED;
}


void quat_rotate(float ax, float ay, float az, GLfloat *rm) {

  float qx[2], qy[3], qz[4], t[4], q[4], wz, xy, xx, yy, zz, wx, wy, xz, yz;


  /* create one quaternion for each axis */
  ax /= 2;
  ay /= 2;
  az /= 2;

  qx[0] = cos(ax);
  qx[1] = sin(ax);
  qy[0] = cos(ay);
  qy[2] = sin(ay);
  qz[0] = cos(az);
  qz[3] = sin(az);

  /* now get t=qx*qy */
  t[0] = qx[0]*qy[0];
  t[1] = qx[1]*qy[0];
  t[2] = qx[0]*qy[2];
  t[3] = qx[1]*qy[2];

  /* next get q=t*qz */
  q[0] = t[0]*qz[0] - t[3]*qz[3];
  q[1] = t[1]*qz[0] + t[2]*qz[3];
  q[2] = t[2]*qz[0] - t[1]*qz[3];
  q[3] = t[0]*qz[3] + t[3]*qz[0];

  /* quaternion -> matrix */
  xx = 2*q[1]*q[1];
  yy = 2*q[2]*q[2];
  zz = 2*q[3]*q[3];
  wx = 2*q[0]*q[1];
  wz = 2*q[0]*q[3];
  xy = 2*q[1]*q[2];
  wy = 2*q[0]*q[2];
  xz = 2*q[1]*q[3];
  yz = 2*q[2]*q[3];

  rm[0] = 1 - yy - zz;
  rm[4] = xy - wz;
  rm[8] = xz + wy;
  rm[12] = 0;

  rm[1] = xy + wz;
  rm[5] = 1 - xx - zz;
  rm[9] = yz - wx;
  rm[13] = 0;

  rm[2] = xz - wy;
  rm[6] = yz + wx;
  rm[10] = 1 - xx - yy;
  rm[14] = 0;

  rm[3] = 0;
  rm[7] = 0;
  rm[11] = 0;
  rm[15] = 1;
}


int particles_get(int id, struct particle_system **pa) {

  struct particle_system *p;


  p = particle_systems;
  while (p != NULL) {
    if (p->id == id) {
      *pa = p;
      return SUCCEEDED;
    }
  }

  return FAILED;
}
